// $Id: Numeric.cpp,v 1.3 2002-05-20 18:28:28 geuzaine Exp $
//
// Copyright (C) 1997 - 2002 C. Geuzaine, J.-F. Remacle
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
// 
// Please report all bugs and problems to "gmsh@geuz.org".

#include "Gmsh.h"
#include "Numeric.h"

// this file should contain only purely numerical routines (that do
// not depend on any Gmsh structures)

double myatan2 (double a, double b){
  if (a == 0.0 && b == 0)
    return 0.0;
  return atan2 (a, b);
}

double myasin(double a){
  if(a<=-1.) return -Pi/2.;
  else if(a>=1.) return Pi/2.;
  else return asin(a);
}

void prodve (double a[3], double b[3], double c[3]){
  c[2] = a[0] * b[1] - a[1] * b[0];
  c[1] = -a[0] * b[2] + a[2] * b[0];
  c[0] = a[1] * b[2] - a[2] * b[1];
}

void prosca (double a[3], double b[3], double *c){
  *c = a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

void norme (double a[3]){
  double mod;
  mod = sqrt (a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
  if (mod == 0.0)
    return;
  a[0] /= mod;
  a[1] /= mod;
  a[2] /= mod;
}

int sys2x2 (double mat[2][2], double b[2], double res[2]){
  double det, ud, norm;
  int i;

  norm = DSQR (mat[0][0]) + DSQR (mat[1][1]) + DSQR (mat[0][1]) + DSQR (mat[1][0]);
  det = mat[0][0] * mat[1][1] - mat[1][0] * mat[0][1];

  // TOLERANCE ! WARNING WARNING
  if (norm == 0.0 || fabs (det) / norm < 1.e-12){
    if(norm) Msg(DEBUG, "Assuming 2x2 matrix is singular (det/norm == %g)", fabs(det)/norm);
    res[0] = res[1] = 0.0 ;
    return 0;
  }
  ud = 1. / det;

  res[0] = b[0] * mat[1][1] - mat[0][1] * b[1];
  res[1] = mat[0][0] * b[1] - mat[1][0] * b[0];

  for (i = 0; i < 2; i++)
    res[i] *= ud;
  return (1);
}

int sys3x3 (double mat[3][3], double b[3], double res[3], double *det){
  double ud;
  int i;

  *det = mat[0][0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) -
    mat[0][1] * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]) +
    mat[0][2] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);

  if (*det == 0.0){
    res[0] = res[1] = res[2] = 0.0 ;
    return (0);
  }

  ud = 1. / (*det);

  res[0] = b[0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) -
    mat[0][1] * (b[1] * mat[2][2] - mat[1][2] * b[2]) +
    mat[0][2] * (b[1] * mat[2][1] - mat[1][1] * b[2]);

  res[1] = mat[0][0] * (b[1] * mat[2][2] - mat[1][2] * b[2]) -
    b[0] * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]) +
    mat[0][2] * (mat[1][0] * b[2] - b[1] * mat[2][0]);

  res[2] = mat[0][0] * (mat[1][1] * b[2] - b[1] * mat[2][1]) -
    mat[0][1] * (mat[1][0] * b[2] - b[1] * mat[2][0]) +
    b[0] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);

  for (i = 0; i < 3; i++)
    res[i] *= ud;
  return (1);

}

int sys3x3_with_tol (double mat[3][3], double b[3], double res[3], double *det){
  int out;
  double norm;

  out = sys3x3(mat,b,res,det);
  norm = 
    DSQR (mat[0][0]) + DSQR (mat[0][1]) + DSQR (mat[0][2]) + 
    DSQR (mat[1][0]) + DSQR (mat[1][1]) + DSQR (mat[1][2]) + 
    DSQR (mat[2][0]) + DSQR (mat[2][1]) + DSQR (mat[2][2]) ;

  // TOLERANCE ! WARNING WARNING
  if (norm == 0.0 || fabs (*det) / norm < 1.e-12){
    if(norm) Msg(DEBUG, "Assuming 3x3 matrix is singular (det/norm == %g)", fabs(*det)/norm);
    res[0] = res[1] = res[2] = 0.0 ;
    return 0;
  }

  return out ;
}

int det3x3 (double mat[3][3], double *det){
  *det = mat[0][0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) -
    mat[0][1] * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]) +
    mat[0][2] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);
  return 1;
}

int inv3x3 (double mat[3][3], double inv[3][3], double *det){
  double ud;

  *det = mat[0][0] * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]) -
    mat[0][1] * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]) +
    mat[0][2] * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);

  if (*det == 0.0)
    return (0);

  ud = 1. / (*det);

  inv[0][0] = ud * (mat[1][1] * mat[2][2] - mat[1][2] * mat[2][1]);
  inv[0][1] = -ud * (mat[1][0] * mat[2][2] - mat[1][2] * mat[2][0]);
  inv[0][2] = ud * (mat[1][0] * mat[2][1] - mat[1][1] * mat[2][0]);
  inv[1][0] = -ud * (mat[0][1] * mat[2][2] - mat[0][2] * mat[2][1]);
  inv[1][1] = ud * (mat[0][0] * mat[2][2] - mat[0][2] * mat[2][0]);
  inv[1][2] = -ud * (mat[0][0] * mat[2][1] - mat[0][1] * mat[2][0]);
  inv[2][0] = ud * (mat[0][1] * mat[1][2] - mat[0][2] * mat[1][1]);
  inv[2][1] = -ud * (mat[0][0] * mat[1][2] - mat[0][2] * mat[1][0]);
  inv[2][2] = ud * (mat[0][0] * mat[1][1] - mat[0][1] * mat[1][0]);
  return (1);

}

double angle_02pi (double A3){
  double DP = 2 * Pi;
  while (A3 > DP || A3 < 0.){
    if (A3 > 0)
      A3 -= DP;
    else
      A3 += DP;
  }
  return A3;
}

double InterpolateIso(double *X, double *Y, double *Z, 
		      double *Val, double V, int I1, int I2, 
		      double *XI, double *YI ,double *ZI){
  if(Val[I1] == Val[I2]){
    *XI = X[I1]; 
    *YI = Y[I1]; 
    *ZI = Z[I1]; 
    return 0;
  }
  else{
    double coef = (V - Val[I1])/(Val[I2]-Val[I1]);
    *XI= coef*(X[I2]-X[I1]) + X[I1];
    *YI= coef*(Y[I2]-Y[I1]) + Y[I1];
    *ZI= coef*(Z[I2]-Z[I1]) + Z[I1];
    return coef;
  }
}

void gradSimplex (double *x, double *y, double *z, double *v, double *grad){
  // p = p1 * (1-u-v-w) + p2 u + p3 v + p4 w

  double mat[3][3];
  double det,b[3];
  mat[0][0] = x[1]-x[0];
  mat[1][0] = x[2]-x[0];
  mat[2][0] = x[3]-x[0];
  mat[0][1] = y[1]-y[0];
  mat[1][1] = y[2]-y[0];
  mat[2][1] = y[3]-y[0];
  mat[0][2] = z[1]-z[0];
  mat[1][2] = z[2]-z[0];
  mat[2][2] = z[3]-z[0];
  b[0] = v[1]-v[0];
  b[1] = v[2]-v[0];
  b[2] = v[3]-v[0];
  sys3x3 (mat, b, grad, &det); 
}
